sw_en = ['hers',
         'except',
         'therefore',
         'across',
         'an',
         'also',
         'toward',
         'n’t',
         '’ve',
         'some',
         'own',
         'five',
         'as',
         'besides',
         'unless',
         'am',
         'thereupon',
         'hence',
         'and',
         'not',
         'becomes',
         'less',
         'together',
         'please',
         'then',
         'last',
         'yourselves',
         'might',
         'of',
         'anything',
         'nor',
         'often',
         'seeming',
         'whereas',
         'one',
         'a',
         'other',
         'throughout',
         'make',
         'top',
         'somewhere',
         'amount',
         'is',
         'out',
         'yours',
         'anyway',
         'formerly',
         'keep',
         'ourselves',
         'since',
         'hundred',
         'yourself',
         'myself',
         'almost',
         'move',
         'otherwise',
         'doing',
         'quite',
         'using',
         'next',
         'behind',
         'onto',
         '‘d',
         'yet',
         'somehow',
         'forty',
         'although',
         'give',
         'wherein',
         'who',
         'thereafter',
         'n‘t',
         '‘s',
         'third',
         'nowhere',
         'among',
         "n't",
         'these',
         'above',
         'seem',
         'see',
         'per',
         'via',
         'towards',
         'wherever',
         '‘ll',
         'show',
         'take',
         'used',
         'without',
         'empty',
         'others',
         "'ll",
         'every',
         'whether',
         'sometime',
         'beforehand',
         'this',
         'whatever',
         'twelve',
         'over',
         'has',
         'so',
         'had',
         'did',
         '‘re',
         '’d',
         'alone',
         'hereafter',
         'four',
         'thru',
         'i',
         'before',
         'afterwards',
         'was',
         'latterly',
         'in',
         'seems',
         'have',
         'seemed',
         'ca',
         'after',
         'indeed',
         'whoever',
         'because',
         'to',
         'someone',
         'it',
         'too',
         'hereupon',
         'during',
         "'d",
         'elsewhere',
         'namely',
         'nevertheless',
         'under',
         'even',
         'more',
         'do',
         'enough',
         'what',
         'back',
         'me',
         'moreover',
         'mine',
         'around',
         're',
         'them',
         'never',
         'with',
         'each',
         'just',
         'which',
         'beside',
         'least',
         'whenever',
         'became',
         'if',
         'us',
         'herself',
         'or',
         'such',
         'whose',
         'full',
         'anyone',
         'whom',
         'here',
         'himself',
         'latter',
         'ten',
         'eleven',
         'much',
         'we',
         'will',
         'hereby',
         'there',
         'neither',
         'nine',
         'something',
         'themselves',
         'would',
         'beyond',
         'fifteen',
         'all',
         'six',
         'up',
         'either',
         'most',
         'upon',
         'at',
         'everyone',
         'none',
         'mostly',
         'the',
         'whither',
         'first',
         "'s",
         'many',
         'may',
         'thus',
         'side',
         '’s',
         'eight',
         'than',
         'put',
         'few',
         'thence',
         'two',
         'rather',
         'three',
         'against',
         'front',
         'perhaps',
         'again',
         'sixty',
         'name',
         '’re',
         'call',
         'former',
         'another',
         'everywhere',
         'still',
         'while',
         'any',
         'else',
         'but',
         'for',
         'get',
         'does',
         'various',
         'whereupon',
         'herein',
         'cannot',
         '’ll',
         'she',
         'below',
         'could',
         "'m",
         'being',
         'ours',
         'twenty',
         "'re",
         'becoming',
         'thereby',
         'ever',
         'they',
         'really',
         'due',
         'further',
         'our',
         'my',
         'between',
         'your',
         'fifty',
         'both',
         'once',
         'were',
         'his',
         'about',
         'are',
         'several',
         'everything',
         'anywhere',
         'from',
         'whole',
         'become',
         'off',
         'on',
         'nothing',
         'now',
         'same',
         'can',
         'you',
         '‘m',
         'itself',
         'however',
         'where',
         'only',
         'within',
         'serious',
         'meanwhile',
         'therein',
         'that',
         'though',
         'been',
         'say',
         'by',
         'its',
         'until',
         'into',
         'whence',
         'part',
         'regarding',
         'him',
         '’m',
         'he',
         "'ve",
         'should',
         'already',
         'go',
         'amongst',
         '‘ve',
         'bottom',
         'their',
         'must',
         'nobody',
         'be',
         'down',
         'those',
         'why',
         'whereafter',
         'no',
         'anyhow',
         'noone',
         'whereby',
         'when',
         'sometimes',
         'her',
         'through',
         'made',
         'along',
         'always',
         'how',
         'done',
         'well',
         'very']

sw_fr = ['doivent',
         'envers',
         'dixième',
         'etais',
         'ah',
         'soixante',
         'l’',
         'suivre',
         'lesquels',
         'specifiques',
         'vos',
         'où',
         'lorsque',
         'ainsi',
         'certain',
         'sera',
         'tres',
         'cette',
         'tente',
         'procedant',
         'as',
         'rendre',
         'vu',
         'dix-sept',
         'meme',
         'soi-meme',
         'cent',
         'sa',
         'le',
         'se',
         'deux',
         'diverses',
         'et',
         'ta',
         'antérieures',
         'étaient',
         "m'",
         'touchant',
         'quant-à-soi',
         'seules',
         'permet',
         'chaque',
         'abord',
         'd’',
         'entre',
         'tenant',
         'mien',
         'pouvait',
         'il',
         "quelqu'un",
         'suis',
         'treize',
         'sont',
         'a',
         'là',
         'comme',
         'aussi',
         'dix-neuf',
         'vé',
         'ceci',
         'chez',
         'vas',
         'parfois',
         'tiennes',
         'j’',
         'mêmes',
         'ceux-là',
         'celui-la',
         'des',
         'plutôt',
         'ouvert',
         'différent',
         "n'",
         'ce',
         'egalement',
         'merci',
         'semblent',
         'cependant',
         'quatrième',
         'juste',
         'ayant',
         'telles',
         'voila',
         "s'",
         'seulement',
         'avons',
         'nouveau',
         'elle-même',
         'encore',
         'hors',
         'sans',
         'tien',
         'outre',
         'partant',
         'très',
         'moi',
         'ai',
         'dessus',
         'parlent',
         'tienne',
         'ces',
         'moins',
         'peux',
         'va',
         'neanmoins',
         'huitième',
         'seuls',
         'ma',
         'differents',
         'te',
         'fait',
         'quoique',
         'cinq',
         'déja',
         'sauf',
         'ni',
         'tant',
         'votres',
         'selon',
         'mais',
         'toutes',
         'ceux-ci',
         'tels',
         'celle-ci',
         'lequel',
         'miens',
         'effet',
         "qu'",
         'via',
         'auquel',
         'doit',
         'debout',
         'quatrièmement',
         'jusque',
         'autres',
         'chacune',
         'cinquantième',
         'celui',
         'devers',
         'pour',
         'devra',
         'compris',
         'dans',
         'cet',
         'etait',
         'dedans',
         'quand',
         'même',
         'etant',
         'avait',
         'onze',
         'dehors',
         "j'",
         'ouste',
         'si',
         'notre',
         'quelles',
         'elles',
         'derrière',
         'sent',
         "d'",
         'gens',
         'chacun',
         'aux',
         'quelque',
         'quant',
         'troisième',
         'environ',
         'aie',
         'peuvent',
         'fais',
         'i',
         's’',
         'spécifique',
         'néanmoins',
         'semblable',
         'avaient',
         'cinquième',
         'suivant',
         'allons',
         'telle',
         'ouverte',
         'hep',
         'différents',
         'préalable',
         'aupres',
         'puis',
         'sixième',
         'est',
         'toute',
         'vais',
         'devant',
         'façon',
         'anterieures',
         'en',
         'car',
         'moi-meme',
         'toi-meme',
         'suffisant',
         'cela',
         'quarante',
         'certes',
         'par',
         'première',
         'suffit',
         'reste',
         'sous',
         'suivantes',
         'différentes',
         'â',
         'stop',
         'soit',
         'quatre-vingt',
         'plus',
         'revoici',
         'vous',
         'anterieure',
         'hormis',
         'douzième',
         'proche',
         'parmi',
         'concernant',
         'exactement',
         'celle-là',
         'celle-la',
         'un',
         'seul',
         'eu',
         'voici',
         'tend',
         'avec',
         'relativement',
         'auraient',
         'sur',
         'deja',
         'près',
         'me',
         'es',
         'troisièmement',
         'donc',
         'surtout',
         'sinon',
         'sienne',
         'avoir',
         'dejà',
         'revoila',
         'étant',
         'vôtre',
         'malgré',
         'seule',
         'une',
         'desormais',
         'douze',
         'dès',
         'parler',
         'trente',
         'y',
         'suivante',
         'attendu',
         'dits',
         'tous',
         'basee',
         'nombreux',
         'ne',
         'pourrais',
         'etaient',
         'or',
         'je',
         'lui',
         'quiconque',
         'leur',
         'antérieur',
         'desquels',
         'da',
         'assez',
         'deuxième',
         'que',
         'lui-meme',
         'serait',
         'à',
         'suivants',
         'hou',
         'nôtres',
         'hem',
         'ou',
         'c’',
         'miennes',
         'ton',
         'vont',
         'combien',
         'certaines',
         'six',
         'ès',
         'trois',
         'jusqu',
         'bas',
         'auxquels',
         'de',
         'm’',
         'diverse',
         'o',
         'au',
         'souvent',
         'na',
         'sept',
         'quelconque',
         'huit',
         'lès',
         'laisser',
         'specifique',
         'elle',
         'mille',
         'duquel',
         'premièrement',
         't’',
         'toi',
         'ceux',
         "l'",
         'pourquoi',
         'maintenant',
         'mon',
         'dire',
         'quelques',
         'cinquantaine',
         'possibles',
         'malgre',
         'nul',
         'divers',
         'lesquelles',
         'seront',
         'eux',
         'dix-huit',
         'tel',
         'lui-même',
         'rend',
         'facon',
         'vôtres',
         'personne',
         'n’',
         'peut',
         'feront',
         'auxquelles',
         'auront',
         'tenir',
         'tu',
         'seize',
         'etre',
         'suit',
         'differentes',
         'mienne',
         'nous',
         'importe',
         'suffisante',
         'pense',
         'afin',
         'semblaient',
         'les',
         'spécifiques',
         'elles-mêmes',
         'quatorze',
         'dite',
         'delà',
         'revoilà',
         'derriere',
         'durant',
         'onzième',
         'sait',
         'hue',
         'différente',
         'septième',
         'après',
         'avais',
         'quel',
         'memes',
         'dit',
         'possible',
         'faisant',
         'hi',
         'qui',
         'siens',
         'la',
         'peu',
         'houp',
         'nos',
         'du',
         'enfin',
         'ho',
         'ô',
         'desquelles',
         'uns',
         'té',
         'pourrait',
         'font',
         'toujours',
         'prealable',
         'semble',
         'certaine',
         'tiens',
         'parce',
         'premier',
         'bat',
         'celles-ci',
         'qu’',
         'notamment',
         'precisement',
         'directement',
         'deuxièmement',
         'directe',
         'quatre',
         'quelle',
         'déjà',
         'cinquante',
         'on',
         'hui',
         'celles-là',
         'dont',
         'elles-memes',
         "c'",
         'moi-même',
         'celle',
         'ci',
         'voilà',
         'restent',
         'differente',
         'aura',
         'ça',
         'celles-la',
         'eh',
         'neuvième',
         'était',
         'être',
         'puisque',
         'ouias',
         'etc',
         'aurait',
         'ils',
         'ouverts',
         'également',
         'lors',
         'avant',
         'nôtre',
         'depuis',
         'plusieurs',
         'unes',
         'étais',
         'apres',
         'antérieure',
         'certains',
         'eux-mêmes',
         'excepté',
         'pendant',
         'longtemps',
         'comment',
         'moindres',
         'vingt',
         'siennes',
         'elle-meme',
         'autrement',
         "t'",
         'dix',
         'désormais',
         'pas',
         'tout',
         'ait',
         'mes',
         'dessous',
         'seraient',
         'autrui',
         'plutot',
         'different',
         'leurs',
         'pres',
         'ses',
         'nombreuses',
         'ha',
         'tellement',
         'restant',
         'relative',
         'toi-même',
         'son',
         'quinze',
         'vers',
         'celles',
         'nous-mêmes',
         'celui-ci',
         'quoi',
         'anterieur',
         'parle',
         'pu',
         'hé',
         'laquelle',
         'celui-là',
         'soi-même',
         'vous-mêmes',
         'ont',
         'allaient',
         'tes',
         'faisaient',
         'alors',
         'maint',
         'soi',
         'retour',
         'quels',
         'précisement',
         'autre',
         'sien',
         'votre']
